function [y_hat]=AnisFovNLM(z,sigma,search_radius,U_radius,disableFov,selfMap,rho,theta,extraPad)
%
%
%  Foveated Nonlocal Means denoising algorithm (ver 2.12, May 15, 2016)
%  Author:  Alessandro Foi, Tampere University of Technology, Finland
% -------------------------------------------------------------------------
%  SYNTAX
%
%     [y_hat] = AnisFovNLM(z,sigma,search_radius,U_radius,disableFov,selfMap,rho,theta,extraPad)
%
%  y_hat :          output denoised image
%
%  z :              input noisy image to be filtered   [REQUIRED]
%  sigma :          standard deviation of the additive white Gaussian noise  [REQUIRED]
%
%  Optional input parameters ---------------------------------------------------------
%  search_radius :  radius of search window     (e.g., 5 <= search_radius <= 13 )
%  U_radius :       radius of patch             (e.g., 1 <= U_radius <= 10)
%  disableFov :     disables foveation when set to 1  (default 0)
%  selfMap :        enables self-map foveation operators when set to 1  (default 1)
%  rho :            elongation of the blur kernels (rho=1 gives circular-symmetric blur kernels)
%  theta :          orientation of the blur kernels relative to the meridian lines
%  extraPad :       extra padding of the blur kernels support
%
%
%
%  This software implements the Foveated Nonlocal Means algorithm
%  proposed in the papers:   (all available online at http://www.cs.tut.fi/~foi/)
%  [1] A. Foi and G. Boracchi, "Foveated self-similarity in nonlocal image filtering",
%      Proc. IS&T/SPIE EI2012 - Human Vision and Electronic Imaging XVII, 8291-32,
%      Burlingame (CA), USA, Jan. 2012.  http://doi.org/10.1117/12.912217
%  [2] A. Foi and G. Boracchi, "Anisotropically Foveated Nonlocal Image Denoising",
%      Proc. IEEE Int. Conf. Image Process. (ICIP 2013), pp. 464-468, Melbourne, Australia, Sep. 2013.
%      http://doi.org/10.1109/ICIP.2013.6738096
%  [3] A. Foi and G. Boracchi, "Foveated nonlocal self-similarity", to appear in Int. J. Computer Vision, 2016.
%      http://doi.org/10.1007/s11263-016-0898-1
%
%
%  By setting disableFov=1, this software can replicate the results of the
%  NLmeansfilter.m  code by J.V. Manjon Herrera & A. Buades (31.07.2008),
%  which implements the NL-means filter proposed by A. Buades, B. Coll and
%  J.M. Morel in the paper "A non-local algorithm for image denoising".
%  http://www.mathworks.com/matlabcentral/fileexchange/13176-non-local-means-filter
%
%%%

disp(' ');
if ~exist('z','var')
    
    disp(' << function AnisFovNLM.m executed without input; showing demo illustrations of foveation operator >>');
    plotFigures=1;
    
    %     kernel_type=1; U_radius=2; selfMap=0; disableFov=0;
    kernel_type=2; U_radius=6; extraPad=5; selfMap=0; disableFov=0; rho=4; theta=pi/8*0;
    
else
    % initialize some variables
    
    % toggles: fovation/windowing, self-map foveation operator
    if ~exist('disableFov','var')||isempty(disableFov)
        disableFov=0;  % use foveation by default
    end
    if ~exist('selfMap','var')||isempty(selfMap)
        selfMap=1;  % use self-map operators by default
    end
    
    if ~disableFov
        disp('Computing Foveated NL-means estimate (using foveated patch distance) ...');
    else
        disp('Computing NL-means estimate (using windowed patch distance) ...');
    end
    plotFigures=0;
    
end



% anisotropy
if ~exist('theta','var')||isempty(theta)
    theta=0.0*pi;  % theta is the angle between the first axis of the elliptical pdf (main axis if rho>1) and the radial (meridian) line between the blur kernel center and the center of the patch
end
if ~exist('rho','var')||isempty(rho)
    rho=4.0;   % if theta=0, this is the tangent-radial ratio: set to 1 for circular-symmetric blur kernel, >1 for radially elongated elliptical blur kernels, <1 for tangentially elongated elliptical blur kernels
end

% By leaving search_radius and U_radius empty, their values are chosen automatically based on the value of sigma (the signal is assumed to be in the standard [0,255] range)
% search-neighborhood and patch radiuses
if ~exist('U_radius','var')||isempty(U_radius)
    sigmaS=[0 10 20 30 50 70];
    U_radiusS=[0 3 5 6 8 9]*~disableFov+[0 2 5 5 6 7]*~~disableFov;
    U_radius=round(interp1(sigmaS,U_radiusS,min(80,max(5,sigma)),'linear','extrap'));
end

if ~exist('kernel_type','var')
    kernel_type=0;
end



%% Windowing k_kernel  (within the Foveated NL-means this kernel is not used for windowing, but to determine the standard-deviation of the blurring kernels used by the foveation operator)
[u2,u1]=meshgrid(-U_radius:U_radius,-U_radius:U_radius);
if kernel_type==0 %% kernel used in the NL-means implementation by J.V. Manjon Herrera & A. Buades
    % used in Fig. 4,5,6,12 of [3]
    ellInftyDistance=max(abs(u1),abs(u2));
    K_tilde=cumsum((2*(U_radius:-1:0)+1).^-2);
    K_tilde=K_tilde([U_radius,U_radius:-1:1]);
    k_kernel=interp1(0:U_radius,K_tilde,ellInftyDistance,'linear','extrap');
elseif kernel_type==1 %% a-la Wertheim
    % used for Fig. 7 of [3], with U_radius=2; selfMap=0;
    l2distance=sqrt(u1.^2+u2.^2);
    k_kernel=1-(l2distance*(0.67568/U_radius/sqrt(2))).^.059079081;
elseif kernel_type==2 %% a-la Wertheim
    %  used for Fig. 9,10,26 of [3], U_radius=6; selfMap=0 or selfMap=1 ;
    l2distance=sqrt(u1.^2+u2.^2);
    k_kernel=1./((l2distance+2.5)).^2.2;
else
    kaiserBeta=8;
    k_kernel=kaiser(U_radius*2+1,kaiserBeta)*kaiser(U_radius*2+1,kaiserBeta)';
    k_kernel=k_kernel.^0.5;
end
k_kernel=k_kernel/sum(k_kernel(:));

numel_k_kernel=numel(k_kernel);



%% Patching and construction of the foveation operator
if disableFov
    %% use windowing instead of foveation
    FovMat=diag(sqrt(k_kernel(:)));  % the foveation operator equivalent to windowing is composed by scaled Dirac impulses, hence in matrix form it is diagonal
    if ~exist('extraPad','var')||isempty(extraPad)
        extraPad=0;  % there is no use in having extraPad>0 if foveation is replaced by simple windowing, since the blur kernels are all Diracs.
    end
else
    %% use foveation instead of windowing
    p=1-exp(-2*pi);
    ell1norm=sqrt(-2*pi/log(1-p))*sqrt(max(k_kernel(:)));  % THIS IS A NUMBER, THE ELL^1 NORM OF ALL THE BLUR KERNELS, SEE (13) IN [1]
    ell2norm2 = k_kernel; % THIS IS A MATRIX OF THE SIZE OF k_kernel  SEE (14) IN [1]
    
    if selfMap
        extraPad=0;    % in case of self-map foveation operator there is no extra padding of the patch domain.
    else
        if ~exist('extraPad','var')||isempty(extraPad)
            extraPad=ceil(3*sqrt(max(rho,1/rho))*ell1norm./(2*sqrt(pi*min(ell2norm2(:)))));    % use extraPad>0 in order to give more room for the support of the blur kernels centered near the boundary of the patch (this is a 3sigma rule using the best guess of the maximum standard deviation varsigma of the blur kernel as ell1norm./(2*sqrt(pi*min(ell2norm2(:))))
        end
    end
    PatchOffset=[0,0];
    allBlurKernels4D=makeFovBlurKernels(ell1norm,ell2norm2,rho,theta,extraPad,PatchOffset); %% construction of blurring blur kernels used by the foveation operator with a given ell1norm and ell2norm2
    FovMat=reshape(allBlurKernels4D,size(allBlurKernels4D,1)*size(allBlurKernels4D,2),size(allBlurKernels4D,3)*size(allBlurKernels4D,4))';  %% foveation operator is written in matrix form
    % allBlurKernels4D=reshape(FovMat',U_radius*2+1+extraPad*2,U_radius*2+1+extraPad*2,U_radius*2+1,U_radius*2+1);  %%  reverse operation
end

if plotFigures
    figure,bar3handle=bar3(k_kernel,1);for jj=1:numel(bar3handle);set(bar3handle(jj),'faceLighting','flat','facecolor',[1 1 1]*0.9),end,title('windowing kernel'),camproj('persp'),axis tight,grid off,set(gcf,'color',[1 1 1]),set(gca,'xtick',[1 2*U_radius+1],'xticklabel',[-U_radius,U_radius]),set(gca,'ytick', [1 2*U_radius+1],'yticklabel',[-U_radius,U_radius]),set(gca,'fontSize',20),camlight(-60,-30),camlight(50,60), material dull,view(40,28); % set(gca,'DataAspectRatio',[1 1 max(k_kernel(:))/U_radius/2]),
    gapWidth=2;
    if ~exist('allBlurKernels4D','var');
        allBlurKernels4D=reshape(FovMat',U_radius*2+1+extraPad*2,U_radius*2+1+extraPad*2,U_radius*2+1,U_radius*2+1);
    end
    makeMosaic(allBlurKernels4D,gapWidth,0.8,0,'blur kernels'); % generate the figure with the mosaic of blur kernels
    [L S R]=svd(FovMat',0); % singular value decomposition of the foveation operator
    
    % optional switching of signs, so to have (without loss of generality) positive center in all left singular functions
    %  diagSign=diag(sign(double(L(round((size(L,1)+1)/2),:)>=0)-0.5));   L=L*diagSign; R=R*diagSign;
    
    allPSFs4D=reshape(FovMat,U_radius*2+1,U_radius*2+1,U_radius*2+1+extraPad*2,U_radius*2+1+extraPad*2);    % point-spread functions
    allLeftSingFun4D=reshape(L,U_radius*2+1+extraPad*2,U_radius*2+1+extraPad*2,U_radius*2+1,U_radius*2+1);    % left singular functions
    allRightSingFun4D=reshape(R,U_radius*2+1,U_radius*2+1,U_radius*2+1,U_radius*2+1);  % right singular functions
    makeMosaic(allPSFs4D,gapWidth,0.8,0,'PSFs');                              % generate the figure with the mosaic of point-spread functions
    makeMosaic(allLeftSingFun4D,gapWidth,1,1,'left singular functions');      % generate the figure with the mosaic of left singular functions
    figure,plotHandle=plot(diag(S));set(get(plotHandle,'parent'),'yscale','log');axis tight;title('singular values');   % plot of singular values
    makeMosaic(allRightSingFun4D,gapWidth,1,1,'right singular functions');    % generate the figure with the mosaic of right singular functions
end

% UcenterIdx=ceil((U_radius*2+1)^2/2);   % index-coordinate of the center of the patch


%% get and verify image size
if ~exist('z','var')
    return
end
[size_z_1 size_z_2 numColorChan]=size(z);
if numColorChan>1
    error('Error: current code works for grayscale images only.')
end

% By leaving search_radius and U_radius empty, their values are chosen automatically based on the value of sigma (the signal is assumed to be in the standard [0,255] range)
% search-neighborhood and patch radiuses
if ~exist('search_radius','var')||isempty(search_radius)
    sigmaS=[0 10 20 30 50 70];
    search_radiusS=[8 8 8 8 8 8]*~disableFov+[5 5 4 4 5 6]*~~disableFov;
    search_radius=round(interp1(sigmaS,search_radiusS,min(80,max(5,sigma)),'linear','extrap'));
end

z_pad=padarray(z,[U_radius+extraPad U_radius+extraPad 0],'symmetric');  % padding (to handle image boundaries without caring about size of patch and blur kernels support)
BIGfov=zeros(numel_k_kernel,size_z_1,size_z_2); % big 3-D array with all foveated patches;
if disableFov
    for x1=1:size_z_1
        for x2=1:size_z_2
            z_x=z_pad(x1:x1+2*(U_radius+extraPad),x2:x2+2*(U_radius+extraPad));  % extract from the padded input image a patch centered at (x1,x2)
            BIGfov(:,x1,x2)=sqrt(k_kernel(:)).*z_x(:);
        end
    end
else
    for x1=1:size_z_1
        for x2=1:size_z_2
            z_x=z_pad(x1:x1+2*(U_radius+extraPad),x2:x2+2*(U_radius+extraPad));  % extract from the padded input image a patch centered at (x1,x2)
            BIGfov(:,x1,x2)=FovMat*z_x(:);
        end
    end
end

if ~disableFov
    disp(' - Filtering with Foveated NL-means ...  ')
else
    disp(' - Filtering with NL-means (windowing, foveation is disabled) ...  ')
end
disp([' . sigma=',num2str(sigma), ', image=',num2str(size_z_2),'x',num2str(size_z_1),', search_radius=',num2str(search_radius),', U_radius=',num2str(U_radius)]);


%% Foveated NL-means.

y_hat=zeros(size_z_1,size_z_2);
BIGfov=permute(BIGfov,[2 3 1]);

timeFirst=now;
timeOld=timeFirst;
loopCount=0;
loopCountPartial=0;
loopTotal=size_z_1*size_z_2;
progressStringEveryNseconds=5.0;
timeLoop=inf;
for x1=1:size_z_1
    for x2=1:size_z_2
        
        % set boundaries of search window
        search1min=max(x1-search_radius,1);  search1max=min(x1+search_radius,size_z_1);  % vertical (left; right)
        search2min=max(x2-search_radius,1);  search2max=min(x2+search_radius,size_z_2);  % horizontal (top; bottom)
        search1center=x1-search1min+1; search2center=x2-search2min+1;  % relative coordinates of [x1,x2]
        
        % compute foveated differences and foveated distance for all foveated patches within the search window
        dFov=0;
        for u=1:numel_k_kernel
            dFov=dFov+(BIGfov(search1min:search1max,search2min:search2max,u)-BIGfov(x1,x2,u)).^2; % compute foveated distance
        end
        
        % visualize foveated distances
        if 0
            if ~exist('axisHandle','var')
                figHandle=figure;
                axisHandle=axes('parent',figHandle);
            end
            image((search2min:search2max)-x2,(search1min:search1max)-x1,dFov,'parent',axisHandle,'CDataMapping','scaled');  % colorbar;
            title(axisHandle,num2str([x1 x2]));
            axis(axisHandle,'equal','tight');
            pause(eps)
        end
        
        w=exp(-dFov/(sigma*sigma));
        w(search1center,search2center,:)=0; wmax=max(w(:)); w(search1center,search2center,:)=wmax+~wmax;  % weight for z(x1,x2) is set equal to the maximum weight (unless all weights are zero, in which case it is set to 1)
        w=w/sum(w(:));
        
        % weighted averaging
        y_hat(x1,x2)=sum(sum(w.*z(search1min:search1max,search2min:search2max)));
        
        
        % progress update string
        loopCount=loopCount+1;
        loopCountPartial=loopCountPartial+1;
        if timeLoop*loopCountPartial>progressStringEveryNseconds;
            timeNow=now;
            timeLoop=86400*(timeNow-timeOld)/loopCountPartial;
            if 86400*(timeNow-timeOld)>progressStringEveryNseconds;
                timeOld=timeNow;
                loopCountPartial=0;
                disp(['      ',num2str(floor(100*loopCount/loopTotal)),' %  (ETA ',num2str(timeLoop*(loopTotal-loopCount)),'s)'])
            end
        end
    end
end
disp(['     ',num2str(100),' %  completed  (',num2str(86400*(now-timeFirst)),'s)'])
%% end of code

